#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include <gst/gst.h>

#include "Emusic.h"
#include "emusic_private.h"


typedef struct _Emusic_Gstreamer_Private Emusic_Gstreamer_Private;
typedef struct  _Emusic_Gsteamer_Eos_Timer_Struct Emusic_Gsteamer_Eos_Timer_Struct;

struct  _Emusic_Gsteamer_Eos_Timer_Struct
  {
    Emusic_Controller *controller;
    void* module;
  };

enum
{
  CDDA_MODE,
  FILE_MODE,
  HTTP_MODE,
  NULL_MODE
};

struct _Emusic_Gstreamer_Private
  {
    Ecore_List *controllers;
    int nb_controllers;
    GstElement *pipeline;
    GstBus *eos_bus;
    Ecore_Timer *eos_timer;
    GstElement *audio;
    double duration;
    unsigned char play : 1;
    unsigned char pause : 1;
    unsigned char stop : 1;
    unsigned char mode;
  };


/* Interface */

static int mod_init(void **module);
static int mod_shutdown(void *module);
static Emusic_Controller *mod_controller_new(void *module, char *name);
static int mod_controller_remove(void *module, Emusic_Controller *controller);
static char *mod_controller_name_get(void *module, Emusic_Controller *controller);
static int mod_controller_name_set(void *module, Emusic_Controller *controller, char *name);
static Emusic_Controller_Mode mod_controller_mode_get(void *module, Emusic_Controller *controller);
static int mod_controller_mode_set(void *module, Emusic_Controller *controller, Emusic_Controller_Mode mode);
static int mod_controller_playlist_new(void *module, Emusic_Controller *controller, char *playlist_name);
static int mod_controller_playlist_add(void *module, Emusic_Controller *controller, char *filename, char *playlist_name);
static int mod_controller_playlist_remove(void *module, Emusic_Controller *controller, char *filename);
static int mod_controller_playlist_remove_id (void *module, Emusic_Controller *controller, int id);
static int mod_controller_playlist_goto(void *module, Emusic_Controller *controller, char *filename);
static int mod_controller_playlist_goto_id (void *module, Emusic_Controller *controller, int id);
static int mod_controller_playlist_destroy(void *module, Emusic_Controller *controller, char *playlist_name);
static int mod_controller_playlist_load(void *module, Emusic_Controller *controller, char *filename);
static int mod_controller_playlist_save(void *module, Emusic_Controller *controller, char *filename);
static Emusic_Playlist *mod_controller_playlist_get(void *module, Emusic_Controller *controller, char *playlist_name);
static int mod_controller_play(void *module, Emusic_Controller *controller);
static int mod_controller_pause(void *module, Emusic_Controller *controller);
static int mod_controller_stop(void *module, Emusic_Controller *controller);
static int mod_controller_next(void *module, Emusic_Controller *controller);
static int mod_controller_prev(void *module, Emusic_Controller *controller);
static double mod_controller_volume_get(void *module, Emusic_Controller *controller);
static int mod_controller_volume_set(void *module, Emusic_Controller *controller, double volume);
static int mod_controller_mute_get(void *module, Emusic_Controller *controller);
static int mod_controller_mute_set(void *module, Emusic_Controller *controller, int mute);
static double mod_controller_length_get(void *module, Emusic_Controller *controller);
static double mod_controller_position_get(void *module, Emusic_Controller *controller);
static int mod_controller_position_set(void *module, Emusic_Controller *controller, double position);

/* Callbacks */
static void _free_controllers_cb(void *data);
static void _playlist_destroy_cb (void *data) ;
static int _eos_timer_fct (void *data);

/* Gstreamer Functions */
static int _pipeline_file_build(void *module);
static int _pipeline_cdda_build(void *module);
static int _pipeline_http_build(void *module);
static int _pipeline_pause (GstElement *pipeline);
static int _pipeline_ready (GstElement *pipeline);
static void _new_decoded_pad_cb (GstElement *decodebin,
                                 GstPad     *pad,
                                 gboolean    last,
                                 gpointer    data);
static void _pipeline_free(void *module);


static Emusic_Module mod =
  {
    mod_init,
    mod_shutdown,
    mod_controller_new,
    mod_controller_remove,
    mod_controller_name_get,
    mod_controller_name_set,
    mod_controller_mode_get,
    mod_controller_mode_set,
    mod_controller_playlist_new,
    mod_controller_playlist_add,
    mod_controller_playlist_remove,
    mod_controller_playlist_remove_id,
    mod_controller_playlist_goto,
    mod_controller_playlist_goto_id,
    mod_controller_playlist_destroy,
    mod_controller_playlist_load,
    mod_controller_playlist_save,
    mod_controller_playlist_get,
    mod_controller_play,
    mod_controller_pause,
    mod_controller_stop,
    mod_controller_next,
    mod_controller_prev,
    mod_controller_volume_get,
    mod_controller_volume_set,
    mod_controller_mute_get,
    mod_controller_mute_set,
    mod_controller_length_get,
    mod_controller_position_get,
    mod_controller_position_set
  };

/* Module Initilisation
   Initialize gst
   TODO : free corectly pipeline and priv struct when creation failed
*/
static int
mod_init(void **module)
{
  GError                   *error;
  Emusic_Gstreamer_Private *priv;

  *module = NULL;
  priv = (Emusic_Gstreamer_Private*) g_new0(Emusic_Gstreamer_Private, 1);
  if (!priv)
    return 0;

  priv->controllers = ecore_list_new();
  if (!priv->controllers)
    {
      g_free(priv);
      return 0;
    }
  priv->nb_controllers = 0;
  priv->play = 0;
  priv->pause = 0;
  priv->stop = 1;
  ecore_list_init(priv->controllers);
  ecore_list_set_free_cb(priv->controllers, _free_controllers_cb);
  g_print("[Emusic] [Gst]  Init\n");

  if (!gst_init_check (NULL, NULL, &error))
    {
      ecore_list_destroy(priv->controllers);
      g_free(priv);
      return 0;
    }

  *module = priv;

  return 1;
}


static int
mod_shutdown(void *module)
{
  Emusic_Gstreamer_Private *priv;

  priv = (Emusic_Gstreamer_Private *) module;

  g_print("[Emusic] [Gst]  Shutdown\n");
  if (!priv)
    goto gst_deinit;

  if (priv->pipeline)
    {
      gst_element_set_state (priv->pipeline, GST_STATE_NULL);
      gst_object_unref (GST_OBJECT (priv->pipeline));
    }
  if (priv->eos_bus)
    gst_object_unref (GST_OBJECT (priv->eos_bus));

  if (priv->controllers)
    ecore_list_destroy(priv->controllers);

  g_free(priv);

gst_deinit:
  gst_deinit ();

  return 1;
}


static Emusic_Controller*
mod_controller_new(void *module, char *name)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Controller *controller;

  priv = (Emusic_Gstreamer_Private*) module;

  controller = (Emusic_Controller *)malloc(sizeof(Emusic_Controller));
  if (!controller)
    return NULL;

  if (!name)
    {
      char buf[256];

      g_snprintf(buf, 256, "controller%d", priv->nb_controllers);
      controller->name = g_strndup(buf, 256);
    }
  else
    controller->name = g_strdup(name);

  controller->mode = EMUSIC_CONTROLLER_MODE_NULL;
  controller->playlists = ecore_list_new();
  if (!controller->playlists)
    {
      g_free(controller->name);
      g_free(controller);
      return NULL;
    }

  ecore_list_init(controller->playlists);
  ecore_list_append(priv->controllers, controller);
  priv->nb_controllers++;

  g_print("[Emusic] [Gst]  nbcontrollers : %d\n", priv->nb_controllers);

  return controller;
}

static char*
mod_controller_name_get(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private *priv;

  priv = (Emusic_Gstreamer_Private*) module;

  return controller->name;
}

static int
mod_controller_name_set(void *module, Emusic_Controller *controller, char *name)
{
  Emusic_Gstreamer_Private *priv;

  priv = (Emusic_Gstreamer_Private *) module;

  controller->name = g_strdup(name);

  return 1;
}

static Emusic_Controller_Mode
mod_controller_mode_get(void *module, Emusic_Controller *controller)

{
  Emusic_Gstreamer_Private *priv;

  priv = (Emusic_Gstreamer_Private*) module;

  return controller->mode;
}

static int
mod_controller_mode_set(void *module, Emusic_Controller *controller, Emusic_Controller_Mode mode)
{
  Emusic_Gstreamer_Private *priv;

  priv = (Emusic_Gstreamer_Private *) module;

  controller->mode = mode;

  g_print("[Emusic] [Gst]  Create Pipeline\n");

  switch (mode)
    {
    case EMUSIC_CONTROLLER_MODE_PLAYER:
      priv->mode = NULL_MODE;
      break;
    case EMUSIC_CONTROLLER_MODE_STREAMER:
      break;
    case EMUSIC_CONTROLLER_MODE_TRANSCODER:
      break;
    default:
      return 0;
    }


  return 1;
}

static int
mod_controller_remove(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private  *priv;
  Emusic_Controller         *c;

  priv = (Emusic_Gstreamer_Private*) module;

  c = ecore_list_goto(priv->controllers, controller);
  if (c != controller)
    return 0;

  priv->nb_controllers--;
  g_print("[Emusic] [Gst]  nbcontrollers : %d\n", priv->nb_controllers);

  return ecore_list_remove_destroy(priv->controllers);
}


static int
mod_controller_playlist_new(void *module, Emusic_Controller *controller, char *playlist_name)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist          *playlist;

  priv = (Emusic_Gstreamer_Private*) module;

  /* FIXME: check the allocations */
  playlist = (Emusic_Playlist*)g_new0(Emusic_Playlist, 1);
  playlist->files = ecore_list_new();
  ecore_list_set_free_cb(playlist->files, _playlist_destroy_cb);
  ecore_list_init(playlist->files);
  playlist->index = 0;
  playlist->name = g_strdup(playlist_name);

  ecore_list_append(controller->playlists, playlist);

  ecore_list_goto(controller->playlists, playlist);

  return 1;
}


static int
mod_controller_playlist_add(void *module, Emusic_Controller *controller, char *filename, char *playlist_name)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist *playlist;
  int id = 0;

  priv = (Emusic_Gstreamer_Private*) module;
  
  for (playlist = ecore_list_first(controller->playlists); playlist; playlist = ecore_list_next(controller->playlists))
  {
    if (!strcmp(playlist_name, playlist->name))
      break;
    id++;
  }
  

  
  playlist = (Emusic_Playlist*)ecore_list_goto_index(controller->playlists, id);
  if (!playlist)
    {
      g_print("[Emusic] [Gst]  playlist doesn't exists !\n");
      return 0;
    }

  g_print("[Emusic] [Gst]   append to playlist %s id %d\n", playlist->name, id); 
    
  ecore_list_append(playlist->files, g_strdup(filename));

  return 1;
}

static int
mod_controller_playlist_remove_id(void *module, Emusic_Controller *controller, int id)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist *playlist;

  priv = (Emusic_Gstreamer_Private*) module;

  playlist = ecore_list_current(controller->playlists);

  ecore_list_goto_index(playlist->files, id);
  ecore_list_remove(playlist->files);

  /* We play index that we remove ==> index-- */
  if (id == playlist->index)
    playlist->index--;

  return 1;
}

static int
mod_controller_playlist_remove(void *module, Emusic_Controller *controller, char *filename)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist *playlist;
  int index;

  priv = (Emusic_Gstreamer_Private*) module;

  playlist = ecore_list_current(controller->playlists);

  ecore_list_goto(playlist->files, filename);
  index = ecore_list_index(playlist->files);
  ecore_list_remove(playlist->files);

  if (index == playlist->index)
    playlist->index--;

  return 1;
}

static int
mod_controller_playlist_goto(void *module, Emusic_Controller *controller, char *filename)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist *playlist;
  int index = 0;
  char *tmp;
  priv = (Emusic_Gstreamer_Private*) module;

  playlist = ecore_list_current(controller->playlists);

  ecore_list_goto_first(playlist->files);
  while( (tmp = ecore_list_next(playlist->files)) )
    {
      if (!strcmp(filename, tmp))
        {
          printf("Found ! index %d\n", index);
          break;
        }
      index++;

    }

  tmp = ecore_list_goto_index(playlist->files, index);
  printf("%s\n", tmp);
  printf("new/old index %d/%d\n", index, playlist->index);
  if (index == playlist->index)
    {
      printf("return same file\n");
      return 0;
    }


  playlist->index = index;

  priv->stop = 1;
  priv->play = 0;
  priv->pause = 0;
  ecore_timer_del(priv->eos_timer);
  priv->eos_timer = NULL;
  return mod_controller_play(module, controller);
}

static int
mod_controller_playlist_goto_id(void *module, Emusic_Controller *controller, int id)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist *playlist;

  priv = (Emusic_Gstreamer_Private*) module;

  playlist = ecore_list_current(controller->playlists);
  
  if (!playlist)
    return 0;
  
  if (id == playlist->index)
    return 1;

  ecore_list_goto_index(playlist->files, id);

  playlist->index = id;

  priv->stop = 1;
  priv->play = 0;
  priv->pause = 0;
  return   mod_controller_play(module, controller);
}

static int
mod_controller_playlist_destroy(void *module, Emusic_Controller *controller, char *playlist_name)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist *playlist;
  int id = 0;
  
  priv = (Emusic_Gstreamer_Private*) module;
  
  g_print("[Emusic] [Gst] destroying playlist_name : %s\n", playlist_name);
  
  if (!controller || !playlist_name)
    return 0;
  
  for (playlist = ecore_list_first(controller->playlists); playlist; playlist = ecore_list_next(controller->playlists))
  {
    if (!strcmp(playlist_name, playlist->name))
      break;
    id++;
  }
  
  playlist = (Emusic_Playlist*)ecore_list_goto_index(controller->playlists, id);
  if (!playlist)
    {
      g_print("[Emusic] [Gst]  playlist doesn't exists !\n");
      return 0;
    }
  
  ecore_list_destroy(playlist->files);
  ecore_list_remove(controller->playlists);

  return 1;
}

static int
mod_controller_playlist_load(void *module, Emusic_Controller *controller, char *filename)
{
  char                      m3u_entry[PATH_MAX];
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist          *playlist = NULL;
  FILE                     *f;

  if (!module || !filename)
    return 0;

  priv = (Emusic_Gstreamer_Private*) module;
  if (ecore_list_is_empty(controller->playlists))
    return 0;

  playlist = ecore_list_current(controller->playlists);
  if (!playlist)
    return 0;

  f = fopen (filename, "rb");
  if (!f)
    return 0;

  while (fgets(m3u_entry, PATH_MAX, f))
    {
      char *c;

      c = m3u_entry + strlen(m3u_entry) - 1;
      if (*c == '\n')
        *c = '\0';

      if (m3u_entry[0] == '#')
        continue;

      printf ("entry: %s\n", m3u_entry);
      ecore_list_append(playlist->files, g_strdup(m3u_entry));
    }

  return 1;
}

static int
mod_controller_playlist_save(void *module, Emusic_Controller *controller, char *filename)
{
  Emusic_Gstreamer_Private* priv;
  Emusic_Playlist          *playlist = NULL;
  gchar                    *file;
  FILE                     *f;

  if (!module || !filename)
    return 0;

  priv = (Emusic_Gstreamer_Private*) module;
  if (ecore_list_is_empty(controller->playlists))
    return 0;

  playlist = ecore_list_current(controller->playlists);
  if (!playlist)
    return 0;

  if (ecore_list_is_empty(playlist->files))
    return 0;

  f = fopen (filename, "wb");
  if (!f)
    return 0;

  ecore_list_goto_first(playlist->files);
  while ( (file = ecore_list_next(playlist->files)) )
    {
      size_t size;

      /* CDDA: we skip */
      if (strstr (file, "cdda://"))
        continue;

      /* FILE: we remove file:// if it exists */
      if (strstr (file, "file://"))
        file += strlen("file://");

      /* HTTP: we store the complete address, so we do nothing */

      size = fwrite(file, 1, strlen(file), f);
      if (size != strlen(file))
        continue;
      size = fwrite("\n", 1, 1, f);
      if (size != 1)
        continue;
    }

  fclose (f);

  return 1;
}

static Emusic_Playlist *
mod_controller_playlist_get(void *module, Emusic_Controller *controller, char *playlist_name)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist          *ep = NULL;

  priv = (Emusic_Gstreamer_Private*) module;

  ecore_list_goto_first(controller->playlists);
  while ( (ep = ecore_list_next(controller->playlists)) )
    {
      g_print("[Emusic] [Gst]  name : %s\n", ep->name);
      if ( ! g_ascii_strcasecmp(ep->name, playlist_name) )
        {
          ecore_list_goto(controller->playlists, ep);
          return ep;
        }
    }

  return NULL;
}


static int
mod_controller_play(void *module, Emusic_Controller *controller)
{
  Emusic_Gsteamer_Eos_Timer_Struct *eos_struct;
  Emusic_Gstreamer_Private         *priv;
  Emusic_Playlist                  *playlist;
  GstElement                       *filesrc;
  gchar                            *file;
  const char                       *device = NULL;
  unsigned int                      track = 1;

  priv = (Emusic_Gstreamer_Private*) module;

  if (!controller->playlists)
    return 0;

  if (!priv->play && priv->stop)
    {
      const char *filename;

      playlist = ecore_list_current(controller->playlists);
      file = (char*) ecore_list_goto_index(playlist->files, playlist->index);
      filename = file;
      if (!file)
        return 0;
      else
        ecore_event_add(controller->signals[EMUSIC_SIGNAL_NEW_FILE], g_strdup(filename), NULL, NULL);

      /* CD Audio */
      if (strstr (file,"cdda://"))
        {

          device = file + strlen ("cdda://");
          if (device[0] == '/')
            {
              char *tmp;

              if ((tmp = strchr (device, (const char)'?')) || (tmp = strchr (device, (const char)'#')))
                {
                  sscanf (tmp + 1,"%d", &track);
                  tmp[0] = '\0';
                }
            }
          else
            {
              device = strdup("/dev/cdrom");
              sscanf (file,"cdda://%d", &track);
            }

          if (priv->mode != CDDA_MODE)
            {
              g_print ("build cdda pipeline\n");
              _pipeline_free(priv);
              if (!(_pipeline_cdda_build (priv)))
                {
                  g_print ("error building CDDA pipeline\n");
                  priv->mode = NULL_MODE;
                  return 0;
                }
              priv->mode = CDDA_MODE;
            }
        }
      /* HTTP */
      else if (strstr (file, "http://"))
        {
          if (priv->mode != HTTP_MODE)
            {
              printf ("build HTTP pipeline \n");
              _pipeline_free(priv);
              if (!(_pipeline_http_build (priv)))
                {
                  priv->mode = NULL_MODE;
                  g_print ("error building HTTP pipeline\n");
                  return 0;
                }
              priv->mode = HTTP_MODE;
            }
        }
      /* Normal media file */
      else
        {


          filename = strstr (file, "file://")
                     ? file + strlen ("file://")
                     : file;

          if (priv->mode != FILE_MODE)
            {
              printf ("build file pipeline \n");
              _pipeline_free(priv);
              if (!(_pipeline_file_build (priv)))
                {
                  printf ("error building File pipeline\n");
                  priv->mode = NULL_MODE;
                  return 0;
                }
              priv->mode = FILE_MODE;
            }
        }

      eos_struct = g_malloc(sizeof( Emusic_Gsteamer_Eos_Timer_Struct));
      eos_struct->module = module;
      eos_struct->controller = controller;

      if (!_pipeline_ready(priv->pipeline))
        return 0;

      switch (priv->mode)
        {
        case FILE_MODE:
          {
            filesrc = gst_bin_get_by_name (GST_BIN (priv->pipeline), "source");
            g_object_set (G_OBJECT (filesrc), "location", filename, NULL);
            break;
          }
        case CDDA_MODE:
          {
            printf("[Emusic] [Gst]  cdda : Track %d Device %s\n", track, device);
            filesrc = gst_bin_get_by_name (GST_BIN (priv->pipeline), "source");
            g_object_set (G_OBJECT (filesrc), "track", track, "device", device, NULL);
            break;
          }
        case HTTP_MODE:
          {
            filesrc = gst_bin_get_by_name (GST_BIN (priv->pipeline), "source");
            g_object_set (G_OBJECT (filesrc), "location", filename, NULL);
            break;
          }
        default:
          g_print ("[Emusic] [Gst]  ON FIRE !\n");
        }

      if (!_pipeline_pause(priv->pipeline))
        return 0;

      /* We get the duration of the audio file */
      if ((priv->mode == FILE_MODE) || (priv->mode == HTTP_MODE))
        {
          GstElement *src;
          GstIterator         *it;
          gpointer             data;

          src = gst_bin_get_by_name (GST_BIN (priv->pipeline), "decodebin");
          it = gst_element_iterate_src_pads (src);
          while (gst_iterator_next (it, &data) == GST_ITERATOR_OK)
            {
              GstPad *pad;
              GstQuery *query;
              gint64 duration = 0;

              pad = GST_PAD (data);
              query = gst_query_new_duration( GST_FORMAT_TIME );
              if ( !gst_pad_query( pad, query ) )
                {
                  g_print( "[Emusic] [Gst]  Can not query the length of the audio pipeline. Set the duration to 0\n" );
                  priv->duration = 0.0;
                  break;
                }
              gst_query_parse_duration( query, NULL, &duration );
              gst_query_unref( query );
              g_print ("[Emusic] [Gst]  duration: %" GST_TIME_FORMAT "\n",
                       GST_TIME_ARGS (duration));
              priv->duration = (double)duration / (double)GST_SECOND;
            }

        }
      else if (priv->mode == CDDA_MODE)
        {
          GstElement *src;
          GstIterator         *it;
          gpointer             data;

          src = gst_bin_get_by_name (GST_BIN (priv->pipeline), "source");
          it = gst_element_iterate_src_pads (src);
          while (gst_iterator_next (it, &data) == GST_ITERATOR_OK)
            {
              GstPad *pad;
              GstQuery *query;
              gint64 duration = 0;

              pad = GST_PAD (data);
              query = gst_query_new_duration( GST_FORMAT_TIME );
              if ( !gst_pad_query( pad, query ) )
                {
                  g_print( "[Emusic] [Gst]  Can not query the length of the audio pipeline. Set the duration to 0\n" );
                  priv->duration = 0.0;
                  break;
                }
              gst_query_parse_duration( query, NULL, &duration );
              gst_query_unref( query );
              g_print ("[Emusic] [Gst]  duration: %" GST_TIME_FORMAT "\n",
                       GST_TIME_ARGS (duration));
              priv->duration = (double)duration / (double)GST_SECOND;
            }
        }

      gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
      priv->play = 1;
      priv->pause = 0;
      priv->stop = 0;
      priv->eos_timer = ecore_timer_add (0.1, _eos_timer_fct, eos_struct);
    }
  else if (priv->pause)
    {
      gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
      priv->play = 1;
      priv->pause = 0;
      priv->stop = 0;
    }

  return 1;
}

static int
mod_controller_pause(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private *priv;

  priv = (Emusic_Gstreamer_Private *) module;

  if (!priv || !priv->pipeline)
    return 0;

  if (priv->play)
    {
      priv->pause = 1;
      priv->play = 0;
      priv->stop = 0;
      if (!_pipeline_pause(priv->pipeline))
        return 0;
    }
  return 0;
}

static int
mod_controller_stop(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private *priv;

  priv = (Emusic_Gstreamer_Private *) module;

  if (!priv ||!priv->pipeline)
    return 0;

  if (priv->play || priv->pause)
    {
      priv->pause = 0;
      priv->play = 0;
      priv->stop = 1;
      if (!_pipeline_ready(priv->pipeline))
        return 0;
    }
  return 0;
}

static int
mod_controller_next(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist          *playlist;

  g_print("[Emusic] [Gst]  mod controller next\n");
  priv = (Emusic_Gstreamer_Private*) module;
  playlist = ecore_list_current(controller->playlists);
  if (!playlist)
    return 0;

  playlist->index++;
  if (!ecore_list_goto_index(playlist->files, playlist->index))
    {
      playlist->index--;
      ecore_list_goto_index(playlist->files, playlist->index);
      return 0;
    }
  priv->stop = 1;
  priv->play = 0;
  priv->pause = 0;
  mod_controller_play(module, controller);

  return 1;
}

static int
mod_controller_prev(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private *priv;
  Emusic_Playlist          *playlist;

  g_print("[Emusic] [Gst]  mod controller prev\n");
  priv = (Emusic_Gstreamer_Private*) module;

  playlist = ecore_list_current(controller->playlists);
  playlist->index--;
  if (!ecore_list_goto_index(playlist->files, playlist->index))
    {
      playlist->index++;
      ecore_list_goto_index(playlist->files, playlist->index);
      return 0;
    }
  priv->stop = 1;
  priv->play = 0;
  priv->pause = 0;
  mod_controller_play(module, controller);

  return 1;
}

static double
mod_controller_volume_get(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private *priv;
  GstElement               *volume;
  double                    vol;

  priv = (Emusic_Gstreamer_Private*) module;

  volume = gst_bin_get_by_name (GST_BIN (priv->pipeline), "volume");
  if (!volume)
    return 0.0;
  g_object_get (G_OBJECT (volume), "volume",
                &vol, NULL);

  return vol;

}

static int
mod_controller_volume_set(void *module, Emusic_Controller *controller, double vol)
{
  Emusic_Gstreamer_Private *priv;
  GstElement* volume;

  priv = (Emusic_Gstreamer_Private*) module;

  if (vol > 10.0)
    vol = 10.0;
  else if (vol < 0.0)
    vol = 0.0;

  volume = gst_bin_get_by_name (GST_BIN (priv->pipeline), "volume");
  if (!volume)
    return 0;

  g_object_set (G_OBJECT (volume), "volume",
                vol, NULL);

  return 1;
}

static int mod_controller_mute_get(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private *priv;
  GstElement *volume;
  gboolean mute;


  if (!module)
    return 0;

  priv = (Emusic_Gstreamer_Private*)module;

  volume = gst_bin_get_by_name (GST_BIN (priv->pipeline), "volume");
  g_object_get (G_OBJECT (volume), "mute",
                &mute, NULL);

  printf("Mute : %d\n", mute);

  return (int)mute;
}

static int mod_controller_mute_set(void *module, Emusic_Controller *controller, int mute)
{
  Emusic_Gstreamer_Private *priv;
  GstElement *volume;

  if (!module)
    return 0;

  priv = (Emusic_Gstreamer_Private*)module;

  volume = gst_bin_get_by_name (GST_BIN (priv->pipeline), "volume");
  g_object_set (G_OBJECT (volume), "mute",
                (gboolean)mute, NULL);

  return 1;
}

static double
mod_controller_length_get(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private* priv;

  if (!module)
    return 0.0;

  priv = (Emusic_Gstreamer_Private*)module;
  return priv->duration;
}

unsigned char
module_open(Emusic_Module **module, void **priv)
{
  if (!module || !priv)
    {
      g_print("[Emusic] [Gst]  Module Adress == NULL\n");
      return 0;
    }

  if (!mod.init(priv))
    {
      g_print("[Emusic] [Gst]  mod.init return 0\n");
      return 0;
    }

  *module = &mod;

  return 1;
}

static double
mod_controller_position_get(void *module, Emusic_Controller *controller)
{
  Emusic_Gstreamer_Private *priv;
  double pos = 0.0;

  if (!module)
    return 0.0;

  priv = (Emusic_Gstreamer_Private*)module;
  if (priv->mode == FILE_MODE)
    {
      GstElement *decodebin;
      GstIterator         *it;
      gpointer             data;

      decodebin = gst_bin_get_by_name (GST_BIN (priv->pipeline), "decodebin");
      it = gst_element_iterate_src_pads (decodebin);
      while (gst_iterator_next (it, &data) == GST_ITERATOR_OK)
        {
          GstPad *pad;
          GstQuery *query;
          gint64 position = 0;

          pad = GST_PAD (data);
          query = gst_query_new_position( GST_FORMAT_TIME );
          if ( !gst_pad_query( pad, query ) )
            {
              g_print( "Gstreamer: can not query the length of the audio pipeline\n" );
              gst_query_unref( query );
              return 0.0;
            }
          gst_query_parse_position( query, NULL, & position );
          gst_query_unref( query );
          pos = (double)position / (double)GST_SECOND;
        }
    }
  else if (priv->mode == CDDA_MODE)
    {
      GstElement *src;
      GstIterator         *it;
      gpointer             data;

      src = gst_bin_get_by_name (GST_BIN (priv->pipeline), "source");
      it = gst_element_iterate_src_pads (src);
      while (gst_iterator_next (it, &data) == GST_ITERATOR_OK)
        {
          GstPad *pad;
          GstQuery *query;
          gint64 position = 0;

          pad = GST_PAD (data);
          query = gst_query_new_position( GST_FORMAT_TIME );
          if ( !gst_pad_query( pad, query ) )
            {
              g_print( "[Emusic] [Gst]  Can not query the position. Set position to 0\n" );
              gst_query_unref( query );
              return 0.0;
            }
          gst_query_parse_position( query, NULL, &position );
          gst_query_unref( query );
          pos = (double)position / (double)GST_SECOND;
        }
    }
  else if (priv->mode == HTTP_MODE)
    {
      return 0.0;
    }
  return pos;

}

static int mod_controller_position_set(void *module, Emusic_Controller *controller, double position)
{
  return 0;
}

void
module_close(Emusic_Module *module)
{
  Emusic_Gstreamer_Private *priv;

  priv = (Emusic_Gstreamer_Private*) module;

  mod.shutdown(priv);
}


/* Callbacks */
static void _free_controllers_cb(void *data)
{
  Emusic_Controller *controller;

  controller = (Emusic_Controller*) data;

  if (!controller)
    return;
  if (controller->name)
    free(controller->name);
  free(controller);
}

static void _playlist_destroy_cb (void *data) 
{
  g_print("[Emusic] [Gst] Destroying \"%s\".\n", (char *)data);
  free(data);
}

static void
_pipeline_free(void *module)
{
  Emusic_Gstreamer_Private *priv;

  if (!module)
    return;

  priv = (Emusic_Gstreamer_Private*) module;

  if (priv->eos_bus)
    gst_object_unref (GST_OBJECT (priv->eos_bus));
  priv->eos_bus = NULL;

  if (priv->pipeline)
    {
      gst_element_set_state (priv->pipeline, GST_STATE_NULL);
      gst_object_unref (GST_OBJECT (priv->pipeline));
      priv->pipeline = NULL;
    }
}

/* Gstreamer Functions */

static int
_pipeline_file_build(void *module)
{
  Emusic_Gstreamer_Private *priv;

  GstElement *src, *dec, *conv, *sink, *vol;
  GstPad *audiopad;

  priv = (Emusic_Gstreamer_Private*) module;

  priv->pipeline = gst_pipeline_new ("pipeline");
  priv->eos_bus = gst_pipeline_get_bus (GST_PIPELINE (priv->pipeline));
  if (!priv->eos_bus)
    g_print("error eos_bus\n");

  src = gst_element_factory_make ("filesrc", "source");
  dec = gst_element_factory_make ("decodebin", "decodebin");
  g_signal_connect (dec, "new-decoded-pad", G_CALLBACK (_new_decoded_pad_cb), priv);
  gst_bin_add_many (GST_BIN (priv->pipeline), src, dec, NULL);
  gst_element_link (src, dec);

  /* create audio output */
  priv->audio = gst_bin_new ("audiobin");

  conv = gst_element_factory_make ("audioconvert", "aconv");
  audiopad = gst_element_get_pad (conv, "sink");
  vol = gst_element_factory_make ("volume", "volume");
  sink = gst_element_factory_make ("autoaudiosink", "sink");
  gst_bin_add_many (GST_BIN (priv->audio), conv, vol, sink, NULL);
  gst_element_link_many (conv, vol, sink, NULL);
  gst_element_add_pad (priv->audio,
                       gst_ghost_pad_new ("sink", audiopad));
  gst_object_unref (audiopad);
  gst_bin_add (GST_BIN (priv->pipeline), priv->audio);

  return 1;
}

static int
_pipeline_http_build(void *module)
{
  Emusic_Gstreamer_Private *priv;

  GstElement *src, *dec, *conv, *sink, *vol;
  GstPad *audiopad;

  priv = (Emusic_Gstreamer_Private*) module;

  priv->pipeline = gst_pipeline_new ("pipeline");
  priv->eos_bus = gst_pipeline_get_bus (GST_PIPELINE (priv->pipeline));
  if (!priv->eos_bus)
    g_print("error eos_bus\n");

  src =gst_element_factory_make ("gnomevfssrc", "source");

  if (!src)
    {
      g_print ("[Emusic] [Gst]  WARNING : try building HTTP source but gnomevsfsrc is not present try with NeonHTTPsrc ...\n");
      src =gst_element_factory_make ("neonhttpsrc", "source");
      if (!src)
        {
          g_print ("[Emusic] [Gst]  WARNING : try building HTTP source but NeonHTTPsrc is not present.\n");
          g_print ("[Emusic] [Gst]  ERROR :  unable to handle HTTP streams, try to install gnomevsfsrc or neonhttpsrc !\n");
          return 0;
        }
      g_print ("[Emusic] [Gst]  HTTP src is neonhttpsrc\n");
    }
  g_print ("[Emusic] [Gst]  HTTP src is gnomevsfsrc\n");

  if ( g_object_class_find_property (G_OBJECT_GET_CLASS (src),"iradio-mode"))
    g_object_set (src, "iradio-mode", TRUE, NULL);

  dec = gst_element_factory_make ("decodebin", "decodebin");
  g_signal_connect (dec, "new-decoded-pad", G_CALLBACK (_new_decoded_pad_cb), priv);
  gst_bin_add_many (GST_BIN (priv->pipeline), src, dec, NULL);
  gst_element_link (src, dec);

  /* create audio output */
  priv->audio = gst_bin_new ("audiobin");

  conv = gst_element_factory_make ("audioconvert", "aconv");
  audiopad = gst_element_get_pad (conv, "sink");
  vol = gst_element_factory_make ("volume", "volume");
  sink = gst_element_factory_make ("autoaudiosink", "sink");
  gst_bin_add_many (GST_BIN (priv->audio), conv, vol, sink, NULL);
  gst_element_link_many (conv, vol, sink, NULL);
  gst_element_add_pad (priv->audio,
                       gst_ghost_pad_new ("sink", audiopad));
  gst_object_unref (audiopad);
  gst_bin_add (GST_BIN (priv->pipeline), priv->audio);

  return 1;
}

static int

_pipeline_cdda_build(void *module)
{
  Emusic_Gstreamer_Private* priv;
  GstElement *src, *conv, *resample, *sink, *vol;

  priv = (Emusic_Gstreamer_Private*) module;

  priv->pipeline = gst_pipeline_new ("pipeline");
  priv->eos_bus = gst_pipeline_get_bus (GST_PIPELINE (priv->pipeline));
  if (!priv->eos_bus)
    g_print("error eos_bus\n");

  src =gst_element_factory_make ("cdparanoiasrc", "source");

  if (!src)
    {
      g_print ("[Emusic] [Gst]  WARNING : try building CDDA source but cdparanoia is not present try using cdiocddasrc ...\n");
      src =gst_element_factory_make ("cdiocddasrc", "source");
      if (!src)
        {
          g_print ("[Emusic] [Gst]  WARNING : try building CDDA source but cdiocddasrc is not present.\n");
          g_print ("[Emusic] [Gst]  ERROR :  unable to handle CDDA streams, try to install cdparanoiasrc or cdiocddasrc !\n");
          return 0;
        }
      g_print ("[Emusic] [Gst]  CDDA src is cdiocddasrc\n");
    }
  else
    g_print ("[Emusic] [Gst]  CDDA src is cdparanoiasrc\n");

  /* create audio output */
  priv->audio = gst_bin_new ("audiobin");

  conv = gst_element_factory_make ("audioconvert", "aconv");
  resample = gst_element_factory_make ("audioresample", "resample");
  vol = gst_element_factory_make ("volume", "volume");
  sink = gst_element_factory_make ("autoaudiosink", "sink");
  gst_bin_add_many (GST_BIN (priv->pipeline), src, conv, resample, vol, sink, NULL);
  gst_element_link_many (src, conv, resample, vol, sink, NULL);

  gst_bin_add (GST_BIN (priv->pipeline), priv->audio);

  return 1;
}

static int
_pipeline_pause (GstElement *pipeline)
{
  GstStateChangeReturn res;

  res = gst_element_set_state ((pipeline), GST_STATE_PAUSED);
  if (res == GST_STATE_CHANGE_FAILURE)
    {
      g_print ("[Emusic] [Gst]  ERROR: could not go to state PAUSED\n");
      return 0;
    }

  res = gst_element_get_state ((pipeline), NULL, NULL, GST_CLOCK_TIME_NONE);
  if (res != GST_STATE_CHANGE_SUCCESS)
    {
      g_print ("[Emusic] [Gst]  ERROR: could not complete the PAUSED state\n");
      return 0;
    }

  return 1;
}

static int
_pipeline_ready (GstElement *pipeline)
{
  GstStateChangeReturn res;

  res = gst_element_set_state ((pipeline), GST_STATE_READY);
  if (res == GST_STATE_CHANGE_FAILURE)
    {
      g_print ("[Emusic] [Gst]  ERROR: could not go to state READY\n");
      return 0;
    }

  res = gst_element_get_state ((pipeline), NULL, NULL, GST_CLOCK_TIME_NONE);
  if (res != GST_STATE_CHANGE_SUCCESS)
    {
      g_print ("[Emusic] [Gst]  ERROR: could not complete the READY state\n");
      return 0;
    }

  return 1;
}

static void
_new_decoded_pad_cb (GstElement *decodebin,
                     GstPad     *pad,
                     gboolean    last,
                     gpointer    data)
{

  GstCaps *caps;
  GstStructure *str;
  GstPad *audiopad;
  Emusic_Gstreamer_Private* priv;

  g_print("[Emusic] [Gst]  new decoded pad\n");
  priv = (Emusic_Gstreamer_Private*) data;
  /* only link once */
  audiopad = gst_element_get_pad (priv->audio, "sink");
  if (GST_PAD_IS_LINKED (audiopad))
    {
      g_object_unref (audiopad);
      return;
    }

  /* check media type */
  caps = gst_pad_get_caps (pad);
  str = gst_caps_get_structure (caps, 0);
  if (!strstr (gst_structure_get_name (str), "audio"))
    {
      gst_caps_unref (caps);
      gst_object_unref (audiopad);
      return;
    }
  gst_caps_unref (caps);

  /* link */

  gst_pad_link (pad, audiopad);

}


int _eos_timer_fct (void *data)
{
  Emusic_Gsteamer_Eos_Timer_Struct *eos_struct;
  Emusic_Gstreamer_Private         *priv;
  Emusic_Controller                *controller;
  GstMessage                       *msg;

  eos_struct = ( Emusic_Gsteamer_Eos_Timer_Struct*) data;
  priv = (Emusic_Gstreamer_Private*) eos_struct->module;
  controller = (Emusic_Controller*) eos_struct->controller;

  while ((msg = gst_bus_poll (priv->eos_bus, GST_MESSAGE_ERROR |
                              GST_MESSAGE_EOS | GST_MESSAGE_TAG |
                              GST_MESSAGE_BUFFERING, 0)))
    {
      
      switch (GST_MESSAGE_TYPE(msg))
        {
        case GST_MESSAGE_ERROR:
          {
            gchar *debug;
            GError *err;

            gst_message_parse_error (msg, &err, &debug);
            g_free (debug);

            g_print ("[Emusic] [Gst]  Error: %s\n", err->message);
            g_error_free (err);

            break;
          }
        case GST_MESSAGE_EOS:
          if (priv->eos_timer)
            {
              g_print("[Emusic] [Gst]  End Of Stream\n");
              ecore_event_add(controller->signals[EMUSIC_SIGNAL_EOS], g_strdup("eos"), NULL, NULL);
              ecore_timer_del(priv->eos_timer);
              priv->eos_timer = NULL;
              if (!mod_controller_next(priv, controller))
                ecore_event_add(controller->signals[EMUSIC_SIGNAL_EOPL], g_strdup("eopl"), NULL, NULL);
            }

          /* TODO */
          /* Gestion EOS */
          break;

        case GST_MESSAGE_TAG:
          {
            GstTagList *new_tags;
            Emusic_Playlist* playlist;
            char* filename = NULL;
            gchar* str = NULL;
            guint integer = 0;
            GDate* date = NULL;
            Emusic_Metadata* metadata;

            g_print("[Emusic] [Gst]  New Tag\n");
            gst_message_parse_tag (msg, &new_tags);

            metadata = (Emusic_Metadata*)g_malloc(sizeof(Emusic_Metadata));

            gst_tag_list_get_string (new_tags, GST_TAG_TITLE, &str);
            if (str)
              metadata->title = g_strdup(str);
            else
              metadata->title = g_strdup("unknown");
            printf("[Emusic] [Gst] [TAG] title : %s\n", metadata->title);
            str = NULL;

            gst_tag_list_get_string (new_tags, GST_TAG_ARTIST, &str);
            if (str)
              metadata->artist = g_strdup(str);
            else
              metadata->artist = g_strdup("unknown");
            printf("[Emusic] [Gst] [TAG] artist : %s\n", metadata->artist);
            str = NULL;

            gst_tag_list_get_string (new_tags, GST_TAG_ALBUM, &str);
            if (str)
              metadata->album = g_strdup(str);
            else
              metadata->album = g_strdup("unknown");
            printf("[Emusic] [Gst] [TAG] album : %s\n", metadata->album);
            str = NULL;

            gst_tag_list_get_uint (new_tags, GST_TAG_TRACK_NUMBER, &integer);
            metadata->track_number =  integer;
            printf("[Emusic] [Gst] [TAG] track_number : %d\n", metadata->track_number);

            gst_tag_list_get_uint (new_tags, GST_TAG_TRACK_COUNT, &integer);
            metadata->number_of_track_on_album =  integer;
            printf("[Emusic] [Gst] [TAG] nb tracks : %d\n", metadata->number_of_track_on_album);

            gst_tag_list_get_string (new_tags, GST_TAG_COMMENT, &str);
            if (str)
              metadata->comment = g_strdup(str);
            else
              metadata->comment = g_strdup("no comment");
            printf("[Emusic] [Gst] [TAG] comment : %s\n", metadata->comment);
            str = NULL;

            gst_tag_list_get_date_index (new_tags, GST_TAG_DATE, 0, &date);
            if (date && g_date_valid(date))
              metadata->year = g_date_get_year(date);
            else
              metadata->year = 0;
            printf("[Emusic] [Gst] [TAG] title : %d\n", metadata->year);

            gst_tag_list_get_string (new_tags, GST_TAG_GENRE, &str);
            if (str)
              metadata->genre = g_strdup(str);
            else
              metadata->genre = g_strdup("unknown");
            printf("[Emusic] [Gst] [TAG] genre : %s\n", metadata->genre);
            str = NULL;

            playlist = ecore_list_current(controller->playlists);
            filename = (char*) ecore_list_goto_index(playlist->files, playlist->index);
            metadata->uri = g_strdup(filename);

            ecore_event_add(controller->signals[EMUSIC_SIGNAL_NEW_TAGS], metadata, NULL, NULL);

            break;
          }

        case GST_MESSAGE_BUFFERING:
          {
            const GstStructure *s;
            gint progress;

            s = gst_message_get_structure (msg);
            if (!gst_structure_get_int (s, "buffer-percent", &progress))
              {
                g_warning ("Could not get value from BUFFERING message");
                break;
              }
            g_print("[Emusic] [Gst]  Progress : %d\n", progress);
            break;
          }
        default:
          break;
        }
      gst_message_unref (msg);
    }
  return 1;
}
